import { isKeyOf, tryParseNumber } from "@arktype/util";
import { writeUnpairableComparatorMessage } from "../../reduce/shared.js";
import { extractDateLiteralSource, isDateLiteral } from "../operand/date.js";
import { parseOperand } from "../operand/operand.js";
export const parseBound = (s, start) => {
    const comparator = shiftComparator(s, start);
    if (s.root.kind === "unit") {
        if (typeof s.root.is === "number") {
            s.unsetRoot();
            return s.reduceLeftBound(s.root.is, comparator);
        }
        if (s.root.is instanceof Date) {
            s.unsetRoot();
            const literal = `d'${s.root.description ?? s.root.is.toISOString()}'`;
            return s.reduceLeftBound(literal, comparator);
        }
    }
    return parseRightBound(s, comparator);
};
export const minComparators = {
    ">": true,
    ">=": true
};
export const maxComparators = {
    "<": true,
    "<=": true
};
export const comparators = {
    ...minComparators,
    ...maxComparators,
    "==": true
};
const oneCharComparators = {
    "<": true,
    ">": true
};
export const comparatorStartChars = {
    "<": 1,
    ">": 1,
    "=": 1
};
const shiftComparator = (s, start) => s.scanner.lookaheadIs("=")
    ? `${start}${s.scanner.shift()}`
    : isKeyOf(start, oneCharComparators)
        ? start
        : s.error(singleEqualsMessage);
export const singleEqualsMessage = `= is not a valid comparator. Use == to check for equality`;
const openLeftBoundToSchema = (leftBound) => ({
    min: isDateLiteral(leftBound.limit)
        ? extractDateLiteralSource(leftBound.limit)
        : leftBound.limit,
    exclusive: leftBound.comparator.length === 1
});
// TODO: allow numeric limits for Dates?
export const parseRightBound = (s, comparator) => {
    // store the node that will be bounded
    const previousRoot = s.unsetRoot();
    const previousScannerIndex = s.scanner.location;
    parseOperand(s);
    // after parsing the next operand, use the locations to get the
    // token from which it was parsed
    const limitToken = s.scanner.sliceChars(previousScannerIndex, s.scanner.location);
    s.setRoot(previousRoot);
    const limit = tryParseNumber(limitToken) ??
        (isDateLiteral(limitToken)
            ? extractDateLiteralSource(limitToken)
            : s.error(writeInvalidLimitMessage(comparator, limitToken, "right")));
    // apply the newly-parsed right bound
    const exclusive = comparator.length === 1;
    // if the comparator is ==, both max and min will be applied
    if (comparator[0] !== ">") {
        s.constrainRoot("max", { max: limit, exclusive });
    }
    if (comparator[0] !== "<") {
        s.constrainRoot("min", { min: limit, exclusive });
    }
    if (!s.branches.leftBound) {
        return;
    }
    // if there's an open left bound, perform additional validation and apply it
    if (!isKeyOf(comparator, maxComparators)) {
        return s.error(writeUnpairableComparatorMessage(comparator));
    }
    s.constrainRoot("min", openLeftBoundToSchema(s.branches.leftBound));
    delete s.branches.leftBound;
};
export const writeInvalidLimitMessage = (comparator, limit, boundKind) => `Comparator ${comparator} must be ${boundKind === "left" ? "preceded" : "followed"} by a corresponding literal (was '${limit}')`;
export const invertedComparators = {
    "<": ">",
    ">": "<",
    "<=": ">=",
    ">=": "<=",
    "==": "=="
};
//# sourceMappingURL=bounds.js.map