import { In, TraversalState, arkKind, builtins, inferred, registry } from "@arktype/schema";
import { CompiledFunction, transform } from "@arktype/util";
import { parseGenericParams } from "./parser/generic.js";
import { bindThis } from "./scope.js";
export const createTypeParser = (scope) => {
    const parser = (...args) => {
        if (args.length === 1) {
            // treat as a simple definition
            return new Type(args[0], scope);
        }
        if (args.length === 2 &&
            typeof args[0] === "string" &&
            args[0][0] === "<" &&
            args[0].at(-1) === ">") {
            // if there are exactly two args, the first of which looks like <${string}>,
            // treat as a generic
            const params = parseGenericParams(args[0].slice(1, -1));
            const def = args[1];
            return validateUninstantiatedGeneric(generic(params, def, scope));
        }
        // otherwise, treat as a tuple expression. technically, this also allows
        // non-expression tuple definitions to be parsed, but it's not a supported
        // part of the API as specified by the associated types
        return new Type(args, scope);
    };
    return parser;
};
export const addArkKind = (value, kind) => Object.defineProperty(value, arkKind, {
    value: kind,
    enumerable: false
});
export const hasArkKind = (value, kind) => value?.[arkKind] === kind;
registry().register(TraversalState, "state");
export class Type extends CompiledFunction {
    definition;
    scope;
    config;
    root;
    condition = "";
    alias;
    allows;
    json;
    constructor(definition, scope) {
        const root = parseTypeRoot(definition, scope);
        super(In, `return true ? { data: ${In} } : { problems: [] } `);
        this.definition = definition;
        this.scope = scope;
        // const state = new ${registry().reference("state")}();
        // const morphs = [];
        // 	for(let i = 0; i < morphs.length; i++) {
        // 	    morphs[i]()
        // 	}
        // 	return state.finalize(${In});
        this.root = root;
        this.allows = root.allows;
        this.config = scope.config;
        this.json = this.root.json;
        this.alias = this.root.alias;
    }
    configure(config) {
        this.config = { ...this.config, ...config };
        return this;
    }
    // TODO: should return out
    from(literal) {
        return literal;
    }
    fromIn(literal) {
        return literal;
    }
    // TODO: Morph intersections, ordering
    and(def) {
        return new Type(this.root.and(parseTypeRoot(def, this.scope)), this.scope);
    }
    or(def) {
        return new Type(this.root.or(parseTypeRoot(def, this.scope)), this.scope);
    }
    morph(morph, outValidator) {
        // TODO: tuple expression for out validator
        outValidator;
        return this;
        // return new Type(
        //     this.root.constrain("morph", morph),
        //     this.scope
        // ) as never
    }
    // TODO: based on below, should maybe narrow morph output if used after
    narrow(def) {
        return new Type(this.root.constrain("predicate", def), this.scope);
    }
    array() {
        return new Type(this.root.array(), this.scope);
    }
    keyof() {
        return new Type(this.root.keyof(), this.scope);
    }
    assert(data) {
        const result = this.call(null, data);
        return result.problems ? result.problems.throw() : result.data;
    }
    // TODO: parse these
    equals(other) {
        return this.root === other.root;
    }
    extends(other) {
        return this.root.extends(other.root);
    }
}
const parseTypeRoot = (def, scope, args) => scope.parseDefinition(def, { args: args ?? bindThis(), baseName: "type" });
export const validateUninstantiatedGeneric = (g) => {
    // the unconstrained instantiation of the generic is not used for now
    // other than to eagerly validate that the def does not contain any errors
    g.scope.parseDefinition(g.definition, 
    // once we support constraints on generic parameters, we'd use
    // the base type here: https://github.com/arktypeio/arktype/issues/796
    {
        baseName: "generic",
        args: transform(g.parameters, ([, name]) => [name, builtins.unknown])
    });
    return g;
};
export const generic = (parameters, definition, scope) => {
    return Object.assign((...args) => {
        const argNodes = transform(parameters, ([i, param]) => [
            param,
            parseTypeRoot(args[i], scope)
        ]);
        const root = parseTypeRoot(definition, scope, argNodes);
        return new Type(root, scope);
    }, {
        [arkKind]: "generic",
        parameters,
        definition,
        scope
        // $ is only needed at compile-time
    });
};
//# sourceMappingURL=type.js.map